function [vip,vipsg,info,viplr,vipsglr,S,L] = sgcenhyb(A,nnodes,opts,show,viplr,vipsglr,infolr)
%SGCENHYB Subgraph centrality computation via hybrid algorithm.
%   node = SGCENHYB(A) identifies the most important node, according to
%   subgraph centrality, of a network whose adjacency matrix is A.
%
%   nodes = SGCENHYB(A,nnodes,opts,show) adds more information.
% 	nnodes: number of nodes to be identified (default 1).
%	opts: additional options, see code.
%	show: if set to 1, displays pictures and results (default 0).
%
%   [nodes values info] = SGCENHYB(A,...) gets more information.
%	nodes: list of nodes.
%	values: approximation of subgraph centrality for each node.
%	info: info(1)=0 means success, info(1)=1 means that the weak
%	      convergence test was satisfied (see code), info(1)=2 means
%	      failure. info(2) contains the number of matrix vector
%	      products performed during the computation.

%   Caterina Fenu & Giuseppe Rodriguez, University of Cagliari, Italy
%   David Martin & Lothar Reichel, Kent State University, U.S.A.
%   Email: kate.fenu@gmail.com, rodriguez@unica.it, 
%          dmarti49@kent.edu, reichel@math.kent.edu
%
%   Last revised Feb 18, 2013

if nargin < 4 || isempty(show), show = 0; end
if nargin < 3 || isempty(opts), opts = struct('empty','empty'); end
if (nargin<2) || isempty(nnodes), nnodes = 1; end

if ~isnumeric(A), error('A must be a numeric matrix.'), end
if ~isequal(A,A'), error('A must be symmetric.'), end

nam = fieldnames(opts);
% additional nodes for weak convergence test
if ~any(strcmpi('sgcincr',nam)), opts.sgcincr = 5; end
% tolerance for weak convergence test in sgcenlr
if ~any(strcmpi('sgctoll',nam)), opts.sgctoll = 1e-3; end
% max number of iterations
if ~any(strcmpi('sgcnmax',nam)), opts.sgcnmax = 100; end
% tolerance for quadrature convergence
if ~any(strcmpi('gausstolq',nam)), opts.gausstolq = 1e-3; end

n = size(A,1);

if nargin < 5
	[viplr vipsglr infolr S L opts.gaussmu] = sgcenlr(A,nnodes,opts,show);
end

info = infolr(1);	% 0 strong conv, 1 weak conv, 2 failure
mvp = infolr(3);

%if ~infolr(1), opts.sgcincr = 0; end	% nodes have been identified by sgcenlr

ei = zeros(n,1);
siz = size(viplr,1);
allsg = zeros(siz,1);
iters = zeros(siz,1);

% for gaussexp
theta = opts.gaussmu;
%opts.gaussreorth = 1;

for i = 1:siz
	ei(viplr(i)) = 1;
	[allsg(i) infog] = gaussexp(A,ei,theta,opts,show);
	iters(i) = infog(2);
	mvp = mvp + infog(2);
	ei(viplr(i)) = 0;
end

[sgs isgs] = sort(allsg,'descend');
vip = viplr(isgs);
vipsg = allsg(isgs);
vip = vip(1:nnodes);
vipsg = vipsg(1:nnodes);

info = [info mvp];

