function [vip vipsgc info S L mu MOV] = sgcenlr(A,nnodes,opts,show)
%SGCENLR Subgraph centrality computation via low-rank approximation.
%   node = SGCENLR(A) identifies the most important node, according to
%   subgraph centrality, of a network whose adjacency matrix is A.
%
%   nodes = SGCENLR(A,nnodes,opts,show) adds more information.
% 	nnodes: number of nodes to be identified (default 1).
%	opts: additional options, see code.
%	show: if set to 1, displays pictures and results (default 0).
%
%   [nodes values info S L mu] = SGCENLR(A,...) gets more information.
%	nodes: list of nodes.
%	values: approximation of subgraph centrality for each node.
%	info: info(1)=0 means success, info(1)=1 means that the weak
%	      convergence test was satisfied (see code), info(1)=2 means
%	      failure. info(2:end) from IRBLEIGS.
%	S: cell array containing all the lists of nodes generated.
%	L: lower bounds for subgraph centrality of each node.
%	mu: largest eigenvalue of A.

%   Caterina Fenu & Giuseppe Rodriguez, University of Cagliari, Italy
%   David Martin & Lothar Reichel, Kent State University, U.S.A.
%   Email: kate.fenu@gmail.com, rodriguez@unica.it, 
%          dmarti49@kent.edu, reichel@math.kent.edu
%
%   Last revised Feb 18, 2013

if (nargin<4) || isempty(show), show = 0; end
if (nargin<3) || isempty(opts), opts = struct('empty','empty'); end
if (nargin<2) || isempty(nnodes), nnodes = 1; end

if ~isnumeric(A), error('A must be a numeric matrix.'), end
if ~isequal(A,A'), error('A must be symmetric.'), end

if ~exist('irbleigs','file')
	error('The IRBLEIGS function is not in the search path.')
end

%	check options
nam = fieldnames(opts);
% max number of iterations
if ~any(strcmpi('sgcmaxn',nam)), opts.sgcmaxn = 300; end
% max number of columns for VA
if ~any(strcmpi('sgcmaxva',nam)), opts.sgcmaxva = 200; end
% additional nodes for weak convergence test
%if ~any(strcmpi('sgcincr',nam)), opts.sgcincr = 0; end
% tolerance for weak convergence test
if ~any(strcmpi('sgctoll',nam)), opts.sgctoll = 0; end
% tau
if ~any(strcmpi('sgctau',nam)), opts.sgctau = 1e-1; end
% min number of incr
if ~any(strcmpi('sgcminkappa',nam)), opts.sgcminkappa = 5; end
% max number of incr
if ~any(strcmpi('sgcmaxkappa',nam)), opts.sgcmaxkappa = 100; end
% sgcshift = 1 for spectrum shift, 0 otherwise
if ~any(strcmpi('sgcshift',nam)), opts.sgcshift = 1; end

%	the following options are for irbleigs
% number of eigenvalues
if ~any(strcmpi('k',nam)), opts.k = 20; end
% blocksize
if ~any(strcmpi('blsz',nam)), opts.blsz = 5; end
% number of blocks
if ~any(strcmpi('nbls',nam)), opts.nbls = ceil(opts.k/opts.blsz+1.1) + 2; end
% 'LE', 'SE', numeric value
if ~any(strcmpi('sigma',nam)), opts.sigma = 'LE'; end
% show intermediate results
if ~any(strcmpi('dispr',nam)), opts.dispr = 0; end
% maximum number of iterations
if ~any(strcmpi('maxit',nam)), opts.maxit = 100; end
% tolerance used for convergence
if ~any(strcmpi('tol',nam)), opts.tol = 1e-3; end

% for internal use
restype = 1;
makemovie = 0; MOV = [];

n = size(A,1);
if nargout>3, keepS = 1; else keepS = 0; end
VA = [];	% eigenvectors database

[V D info] = irbleigs(A,opts);
if info(1)
	warning('sgcenlr:irbleigs','IRBLEIGS did not converge.')
end
infosave = info;
lam = diag(D);
[lam vind] = sort(lam,'descend');	% are the eigenvalues already sorted?
V = V(:,vind);
numeig = opts.k;
tsol = zeros(n,1);	% originally, the true solution

if opts.sgcshift	% spectrum shift
	mu = lam(1);
else
	mu = 0;
end

if show
	allam = lam;
	figure(show+1)
	subplot(1,2,1)
	set(gca,'fontsize',12)
	spy(A)
	title('Adjacency matrix', 'fontsize',14)
	subplot(1,2,2)
	set(gca,'fontsize',12)
	plot(lam,'o-')
	axis square
	axis tight
	grid
	title('Computed spectrum', 'fontsize',14)

	ylv = [min(tsol) max(tsol)+eps];
	wid = ylv(2)-ylv(1);
	ylv = [max(0,ylv(1)-wid/2) ylv(2)+wid/2];
	figure(show)
	if any(tsol)
		plot(tsol,'r')
	else
		plot(0);
	end
	xlim([1 n])
	ylim(ylv)
	set(gca,'fontsize',12)
	ht = text(1,1,'');
end

L = zeros(n,1);
vsum = zeros(n,1);
if keepS, S = cell(numeig,1); end

N = 0;
realN = 0;
restart = 0;
flag = 1;
lis = [];
while flag && (realN<min(opts.sgcmaxn,n)) && (N<numeig)
	N = N+1;
	realN = realN+1;
	elam = exp(lam(N)-mu);
	v = V(:,N).^2;
	vsum = vsum + v;
	L = L + elam*v;
	U = L + elam*(1-vsum);
	[Ls indx] = sort(L,'descend');
	Lmax = Ls(nnodes);		% Lmax = max(L);
	lis = find(U >= Lmax);
	if keepS, S{realN} = lis; end

	% exit when the nodes are identified or lower bounds stagnate
	flag = (numel(lis)>nnodes) && (elam/n>opts.sgctoll*Lmax);

	if show
		%fprintf('N=%d, Lmax=%.2e, #S=%d\n', realN, Lmax, numel(lis))
		figure(show)
		hold on
		if any(tsol)
			hh = plot([L U tsol]);
			set(hh(3),'linewidth',1.5)
		else
			plot([L U]);
			ylv(1) = min(ylv(1),1.5*min(L));
			ylv(2) = max(ylv(2),1.5*max(L));
			ylim(ylv)
		end
		title(sprintf('N=%d - #S=%d\n', realN, numel(lis)), ...
			'fontsize',14)
		hold off
		drawnow

		if makemovie
			delete(ht)
			ypos = .8 * (ylv(2)-ylv(1)) + ylv(1);
			ht = text(50,ypos,sprintf('N=%d - #S=%d\n', ...
				realN, numel(lis)), 'fontsize',16);
			MOV(N) = getframe;
		end
	end

	if N == numeig		% restart
		if isfield(opts,'eigvec'), opts = rmfield(opts,'eigvec'); end
		if isfield(opts,'sigma'), opts = rmfield(opts,'sigma'); end
		switch restype
			case 1
				VA = [VA V];
				opts.eigvec = VA;
				[V D info] = irbleigs(A,opts);
				if info(1)
					warning('sgcenlr:irbleigs',...
						'IRBLEIGS did not converge.')
				end
				infosave = infosave + info;
				lam = diag(D);
				[lam vind] = sort(lam,'descend');
				V = V(:,vind);
				if keepS, S = [S; cell(numeig,1)]; end
				restart = restart + 1;
				N = 0;
				% switch to other restart if VA is too large
				if size(VA,2) >= opts.sgcmaxva,
					restype = 2;
				end
			case 2
				opts.sigma = lam(N);
				[V D info] = irbleigs(A,opts);
				if info(1)
					warning('sgcenlr:irbleigs',...
						'IRBLEIGS did not converge.')
				end
				infosave = infosave + info;
				lam = diag(D);
				[lam vind] = sort(lam,'descend');
				V = V(:,vind);
				[ll ill] = min(abs(lam-opts.sigma));
				if keepS, S = [S; cell(numeig-ill,1)]; end
				restart = restart + 1;
				N = ill;
			otherwise
				error('wrong restart')
		end
		if show
			allam = [allam; lam(N+1:numeig)];
			figure(show+1)
			subplot(1,2,2)
			set(gca,'fontsize',12)
			plot(allam,'o-')
			axis square
			axis tight
			grid
			title('Computed spectrum', 'fontsize',14)
			drawnow
		end
	end
end

if keepS, S = S(1:realN); end

if flag
	warning('sgcenlr:tooFewEigenvales',...
		'Too few eigenvalues to identify VIP nodes.')
	info = [2 infosave];
	nvip = numel(lis);
else
	if numel(lis) > nnodes
		ivec = find((Lmax-Ls(nnodes+1:end))<opts.sgctau*Lmax);
		incr = min(length(ivec),opts.sgcmaxkappa);
		if incr < opts.sgcminkappa, incr = opts.sgcminkappa; end
		info = [1 infosave];
	else
		incr = 0;
		info = [0 infosave];
	end
	nvip = nnodes+incr;
end

vip = indx(1:nvip);
vipsgc = Ls(1:nvip);

if isfinite(exp(mu))
	vipsgc = exp(mu)*vipsgc;
else
	warning('sgcenlr:scaledCentrality',...
		'Subgraph centrality is scaled by exp(mu).')
end

if show
	switch info(1)
		case 0
			fprintf('Strong convergence test was satisfied.\n')
		case 1
			fprintf('Weak convergence test was satisfied.\n')
		case 2
			fprintf('No convergence test was satisfied.\n')
	end
	fprintf('N=%d - %d restarts - %d VIP node',realN,restart,nvip)
	if nvip > 1
		fprintf('s identified:\n')
	else
		fprintf(' identified:\n')
	end
	fprintf('%8d (%.2e)\n',[vip vipsgc]')
	figure(show)
	hold on
	plot(vip,vipsgc,'r.','markersize',20)
	hold off
end

