function [vip vipsg info allsg iters] = stcongauss(A,nnodes,opts,show)
%STCONGAUSS Starting convenience computation via Gauss quadrature.
%   node = STCONGAUSS(A) identifies the most important node, according to
%   starting convenience, of a network whose adjacency matrix is A.
%
%   nodes = STCONGAUSS(A,nnodes,opts,show) adds more information.
% 	nnodes: number of nodes to be identified (default 1).
%	opts: additional options, see code.
%	show: if set to 1, displays pictures and results (default 0).
%
%   [nodes values info] = STCONGAUSS(A,...) gets more information.
%	nodes: list of nodes.
%	values: approximation of starting convenience for each node.
%	info: info(1)=0 means success, info(1)=1 means that a convergence test
%	      was not satisfied (see code). info(2) contains the number of
%	      matrix vector products performed during the computation.

%   Caterina Fenu & Giuseppe Rodriguez, University of Cagliari, Italy
%   David Martin & Lothar Reichel, Kent State University, U.S.A.
%   Email: kate.fenu@gmail.com, rodriguez@unica.it, 
%          dmarti49@kent.edu, reichel@math.kent.edu
%
%   Last revised Feb 18, 2013

if nargin < 4 || isempty(show), show = 0; end
if nargin < 3 || isempty(opts), opts = struct('empty','empty'); end

if ~isnumeric(A), error('A must be a numeric matrix.'), end
if ~isequal(A,A'), error('A must be symmetric.'), end

nam = fieldnames(opts);
% tolerance for quadrature convergence
if ~any(strcmpi('gausstolq',nam)), opts.gausstolq = 1e-3; end
% max number of iterations
if ~any(strcmpi('gaussmaxn',nam)), opts.gaussmaxn = 200; end
% spectrum shift for gaussexp
if ~any(strcmpi('gaussmu',nam)), opts.gaussmu = eigs(double(A),1,'LA'); end

n = size(A,1);

allsg = zeros(n,1);
iters = zeros(n,1);
c = ones(n,1);
mvp = 0;
info = 0;
theta = opts.gaussmu;

np3 = n+3;  nm1 = n-1;
[den infog] = gaussexp(A,c/sqrt(n),theta,opts,show);
info = info || infog(1);

y = [2;ones(n-1,1)]/sqrt(np3);
z = [0;ones(n-1,1)]/sqrt(nm1);
[allsg1 infog] = gaussexp(A,y,theta,opts,show);
info = info || infog(1);
iters(1) = infog(2);
mvp = mvp + infog(2);
[allsg2 infog] = gaussexp(A,z,theta,opts,show);
info = info || infog(1);
iters(1) = iters(1) + infog(2);
mvp = mvp + infog(2);
allsg(1) = np3*allsg1 - nm1*allsg2;
for i = 2:n
	y([i-1 i]) = y([i i-1]);
	z([i-1 i]) = z([i i-1]);
	[allsg1 infog] = gaussexp(A,y,theta,opts,show);
	info = info || infog(1);
	iters(i) = infog(2);
	mvp = mvp + infog(2);
	[allsg2 infog] = gaussexp(A,z,theta,opts,show);
	info = info || infog(1);
	iters(i) = iters(1) + infog(2);
	mvp = mvp + infog(2);
	allsg(i) = np3*allsg1 - nm1*allsg2;
end
allsg = allsg/(4*den);

[vipsg vip] = sort(allsg,'descend');

vip = vip(1:nnodes);
vipsg = vipsg(1:nnodes);

if any(iters) >= opts.gaussmaxn
	info = 1;
end
info = [info mvp];

